/**
 * Define api of apeman-api-fs
 * @function define
 * @param {Object} [options] - Optional settings
 * @param {boolean} [options.logging] - Enable logger
 * @param {string} [options.root=process.cwd()] - Root directory path
 * @returns {Object} - API configuration
 */
'use strict'

const co = require('co')
const { mkdirpAsync, existsAsync, statAsync } = require('asfs')
const filemode = require('filemode')
const filelink = require('filelink')
const aglob = require('aglob')
const aslogger = require('aslogger')
const filecopy = require('filecopy')
const apemanApiPath = require('apeman-api-path')
const load = require('apeman-api-client/load')
const spec = require('./spec')

let modeString = (mode) => mode.toString(8)

/** @lends define */
function define (options = {}) {
  let {
    logging = false,
    root = process.cwd()
  } = options
  let logger = aslogger({ disabled: !logging })
  let pathApi = apemanApiPath({ root, logging: false })

  return {
    /**
     * Test the reachability of the api.
     * @method ping
     * @param {string} [pong='pong'] - Pong message to return
     * @returns {string} - Pong message
     */
    ping (pong = 'pong') {
      return co(function * () {
        logger.trace('pong:', pong)
        return pong
      })
    },

    /**
     * Expand glob filename patterns
     * @method glob
     * @param {string} pattern - Glob filename pattern
     * @param {Object} [options={}] - Optional settings
     * @returns {Promise}
     */
    glob (pattern, options = {}) {
      const s = this
      let args = [ ...arguments ]
      let { cwd = process.cwd() } = options
      return co(function * () {
        let path = yield load(pathApi, { $$ctx: s })
        let filenames = yield aglob(...args)
        for (let i = 0; i < filenames.length; i++) {
          let filename = filenames[ i ]
          filenames[ i ] = yield path.resolve(filename)
        }
        return filenames
      })
    },

    /**
     * Check file exists
     * @method exists
     * @param filename
     * @returns {Promise}
     */
    exists (filename) {
      const s = this
      return co(function * () {
        let path = yield load(pathApi, { $$ctx: s })
        let resolved = yield path.resolve(filename)
        return yield existsAsync(resolved)
      })
    },

    /**
     * Get file state
     * @method stat
     * @param filename
     * @returns {Promise}
     */
    stat (filename) {
      const s = this
      return co(function * () {
        let path = yield load(pathApi, { $$ctx: s })
        let resolved = yield path.resolve(filename)
        return yield statAsync(resolved)
      })
    },

    /**
     * Make directories
     * @method mkdir
     * @param dirname
     * @returns {Promise}
     */
    mkdir (dirname) {
      const s = this
      return co(function * () {
        let result = yield s.mkdirAll([ dirname ])
        return !!result[ 0 ]
      })
    },

    /**
     * Make multiple directories at once.
     * @method mkdirAll
     * @param dirnames
     * @returns {Promise}
     */
    mkdirAll (dirnames) {
      const s = this
      return co(function * () {
        let path = yield load(pathApi, { $$ctx: s })
        let results = []
        for (let dirname of dirnames) {
          let resolved = yield path.resolve(dirname)
          let exists = yield s.exists(resolved)
          if (exists) {
            continue
          }
          yield mkdirpAsync(resolved)
          logger.debug(`Directory generated: ${dirname}`)
          results.push(dirname)
        }
        return results
      })
    },

    /**
     * Change file permission
     * @method chmod
     * @param {string} pattern - Glob filename pattern
     * @param {string} mode
     * @returns {Promise}
     */
    chmod (pattern, mode) {
      const s = this
      return s.chmodAll({ [pattern]: mode })
    },

    /**
     * Change file permissions at once
     * @method chmodAll
     * @param {Object} permissions
     * @returns {Promise}
     */
    chmodAll (permissions) {
      const s = this
      return co(function * () {
        let results = []
        for (let pattern of Object.keys(permissions)) {
          let mode = permissions[ pattern ]
          let filenames = yield s.glob(pattern)
          for (let filename of filenames) {
            let changed = yield filemode(filename, mode, {})
            Object.keys(changed).forEach((filename) => {
              let { from, to } = changed[ filename ]
              logger.debug(`Permission changed: ${filename} ${modeString(from)} to ${modeString(to)}`)
              results.push(filename)
            })
          }
        }
        return results
      })
    },

    /**
     * Copy a file
     * @method copy
     * @param {string} src - Source file name
     * @param {string} dest - Destination file name
     * @returns {Promise}
     */
    copy (src, dest) {
      const s = this
      return s.copyAll({ [src]: dest })
    },

    /**
     * Copy multiple files at once
     * @method copyAll
     * @param {Object.<string, string>} config - Copy configuration
     * @returns {Promise}
     */
    copyAll (config) {
      const s = this
      return co(function * () {
        let path = yield load(pathApi, { $$ctx: s })
        let results = []
        for (let src of Object.keys(config)) {
          let dest = config[ src ]
          src = yield path.resolve(src)
          dest = yield path.resolve(dest)
          let copied = yield filecopy(src, dest, {
            mkdirp: true
          })
          let changed = Object.keys(copied).shift()
          if (changed) {
            logger.debug(`File copied: ${src} -> ${dest}`)
            results.push(changed)
          }
        }
        return results
      })
    },

    /**
     * Create symlink
     * @method symlink
     * @param {string} src - Source file name
     * @param {string} dest - Destination file name
     * @param {Object} [options={}] - Optional settings
     * @returns {Promise}
     */
    symlink (src, dest, options = {}) {
      const s = this
      return s.symlinkAll({ [src]: dest }, options)
    },

    /**
     * Create multiple symlinks at once
     * @method symlinkAll
     * @param {Object.<string, string>} config - Link configuration
     * @param {Object} [options={}] - Optional settings
     * @returns {Promise}
     */
    symlinkAll (config, options = {}) {
      const s = this
      return co(function * () {
        let path = yield load(pathApi, { $$ctx: s })
        let results = []
        for (let src of Object.keys(config)) {
          let dest = config[ src ]
          src = yield path.resolve(src)
          let exists = yield s.exists(src)
          if (!exists) {
            console.warn(`Symlink src "${src}" not exists`)
            continue
          }
          dest = yield path.resolve(dest)
          let changed = yield filelink(src, dest, {
            type: 'symlink',
            mkdirp: true,
            force: true
          })
          if (changed) {
            logger.debug(`File linked: ${src} -> ${dest}`)
            results.push({ src, dest })
          }
        }
        return results
      })
    },

    /**
     * A hook which runs before method invoked
     * @param {string} methodName - Name of method to invoke
     * @param {Array} params - Prams for invocation
     * @returns {Promise}
     */
    $before (methodName, params) {
      const s = this
      let { state } = s
      return co(function * () {
        state.invokedAt = new Date()
        logger.info(`${methodName} started...`)
      })
    },
    /**
     * A hook which runs after method invoked
     * @param {string} methodName - Name of method to invoke
     * @param {Array} params - Prams for invocation
     * @param {?*} returns - Return values
     * @returns {Promise}
     */
    $after (methodName, params, returns) {
      const s = this
      let { state } = s
      return co(function * () {
        let took = new Date() - state.invokedAt
        logger.info(`...${methodName} done! (${took}ms)`)
      })
    },
    /**
     * Method specifications
     */
    $spec: spec
  }
}

module.exports = define
